#ifndef COYOTE_COYOTEENV_HPP
#define COYOTE_COYOTEENV_HPP

#include <thread>

#include "field.hpp"
#include "tsqueue.hpp"

namespace coyote {

    using Task = std::function<void()>;
    // first is the user-provided prio, second is the iso8601 time to ensure chronological
    // proceeding
    using TaskPrio = std::tuple<int, std::string>;
    struct TaskPrioCompare {
        bool operator()(const TaskPrio& a, const TaskPrio& b) const {
            if (std::get<0>(a) < std::get<0>(b))
                return true;
            if (std::get<0>(a) > std::get<0>(b))
                return false;
            // smaller datetime has higher priority!
            return std::get<1>(a) > std::get<1>(b);
        }
    };
    using TaskQueue = TSQueue<Task, TaskPrio, TaskPrioCompare>;

    class CoyoteEnv {
        friend void coyote_init(std::string group_name, uint nthreads);

      public:
        CoyoteEnv(std::string group_name = "", uint nthreads = 1);

        static CoyoteEnv& getInstance();

        MPI_Comm get_group_comm() const;
        int get_yac_instance_id() const;
        std::string start_datetime();
        std::string end_datetime();
        void add_field(std::unique_ptr<Field> field);
        void ensure_def_comps();
        void ensure_sync_def();
        void ensure_enddef();
        void run();
        void add_task(std::function<void()> task, int prio = 0);

      private:
        static std::unique_ptr<CoyoteEnv> _instance;
        int yac_instance_id;
        MPI_Comm group_comm;
        bool state_def_comps, state_sync_def, state_enddef;
        std::thread put_thread, get_thread;
        TaskQueue _task_queue;
        std::vector<std::unique_ptr<Field>> _fields;
        bool _finalize_mpi = false;
        uint _nthreads     = 1;
    };

} // namespace coyote

#endif
